
 
from distask.locks.base import BaseLock
import logging, os, time
from kazoo.client import KazooClient
from kazoo.recipe.lock import Lock
 
class ZKLock(BaseLock):
    def __init__(self, hosts=None, name="distask_zklock", timeout=10):
        """
        :param hosts: zookeeper主机地址
        :param name: 分布式锁名称
        :param logger: 日志对象
        :param timeout: 连接超时
        """
        self._client = None
        self._lock = None
 
        # 创建客户端对象并初始化连接
        try:
            self._client = KazooClient(hosts=hosts, timeout=timeout)
            self._client.start(timeout=timeout)
        except Exception as e:
            logging.error('Create KazooClient Failed! Exception:{}'.format(e))
            raise
 
        # 创建Lock对象
        try:
            lock_path = os.path.join("/", "locks", name)
            self._lock = Lock(self._client, lock_path)
        except Exception as e:
            logging.error('Create Lock Failed! Exception: %s'.format(e))
            raise
 
    def __del__(self):
        """
        :return:
        """
        self.unlock()
        if self._client:
            self._client.stop()
            self._client = None
 
    def lock(self, blocking=True, timeout=None):
        """
        """
        if self._lock is None:
            return False
 
        try:
            return self._lock.acquire(blocking=blocking, timeout=timeout)
        except Exception as e:
            logging.error('Acquire lock failed! Exception: %s'.format(e))
            return False
 
    def unlock(self):
        """
        """
        if self._lock is not None:
            self._lock.release()
            logging.info('Release Lock')



from threading import Event, Thread

from distask.schedulers.base import Scheduler


class BackgroundScheduler(Scheduler):

    _thread = None
    _daemon = True

    def start(self, *args, **kwargs):
        if self._event is None or self._event.is_set():
            self._event = Event()

        Scheduler.start(self, ready=False)
        self._thread = Thread(target=self._main_loop, name='distask scheduler')
        self._thread.daemon = self._daemon
        self._thread.start()

    def shutdown(self, *args, **kwargs):
        super(BackgroundScheduler, self).shutdown(*args, **kwargs)
        # self._thread.join()
        # del self._thread



from datetime import date, timedelta, tzinfo
import datetime
import re
import time
from calendar import timegm
from pytz import FixedOffset, timezone, utc

import string
from inspect import isclass, ismethod
from functools import partial
import six

RETRY_DELAY = 10
        
def time_now():
    return int(time.time())

def micro_now():
    return int(time.time() * 1000)

# 2050年1月1日
def micro_max():
    return int(2524579200_000)

try:
    from threading import TIMEOUT_MAX
except ImportError:
    TIMEOUT_MAX = 4294967 

def safeint(text):
    if text is not None:
        return int(text)


def datetime_ceil(dateval):
    """
    Rounds the given datetime object upwards.

    :type dateval: datetime

    """
    if dateval.microsecond > 0:
        return dateval + timedelta(seconds=1, microseconds=-dateval.microsecond)
    return dateval



_DATE_REGEX = re.compile(
    r'(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})'
    r'(?:[ T](?P<hour>\d{1,2}):(?P<minute>\d{1,2}):(?P<second>\d{1,2})'
    r'(?:\.(?P<microsecond>\d{1,6}))?'
    r'(?P<timezone>Z|[+-]\d\d:\d\d)?)?$')

def convert_to_datetime(input, tz, arg_name):
    """
    Converts the given object to a timezone aware datetime object.

    If a timezone aware datetime object is passed, it is returned unmodified.
    If a native datetime object is passed, it is given the specified timezone.
    If the input is a string, it is parsed as a datetime with the given timezone.

    Date strings are accepted in three different forms: date only (Y-m-d), date with time
    (Y-m-d H:M:S) or with date+time with microseconds (Y-m-d H:M:S.micro). Additionally you can
    override the time zone by giving a specific offset in the format specified by ISO 8601:
    Z (UTC), +HH:MM or -HH:MM.

    :param str|datetime input: the datetime or string to convert to a timezone aware datetime
    :param datetime.tzinfo tz: timezone to interpret ``input`` in
    :param str arg_name: the name of the argument (used in an error message)
    :rtype: datetime

    """
    if input is None:
        return
    elif isinstance(input, datetime):
        datetime_ = input
    elif isinstance(input, date):
        datetime_ = datetime.combine(input, time())
    elif isinstance(input, six.string_types):
        m = _DATE_REGEX.match(input)
        if not m:
            raise ValueError('Invalid date string')

        values = m.groupdict()
        tzname = values.pop('timezone')
        if tzname == 'Z':
            tz = utc
        elif tzname:
            hours, minutes = (int(x) for x in tzname[1:].split(':'))
            sign = 1 if tzname[0] == '+' else -1
            tz = FixedOffset(sign * (hours * 60 + minutes))

        values = {k: int(v or 0) for k, v in values.items()}
        datetime_ = datetime(**values)
    else:
        raise TypeError('Unsupported type for %s: %s' % (arg_name, input.__class__.__name__))

    if datetime_.tzinfo is not None:
        return datetime_
    if tz is None:
        raise ValueError(
            'The "tz" argument must be specified if %s has no timezone information' % arg_name)
    if isinstance(tz, six.string_types):
        tz = timezone(tz)

    try:
        return tz.localize(datetime_, is_dst=None)
    except AttributeError:
        raise TypeError(
            'Only pytz timezones are supported (need the localize() and normalize() methods)')


def datetime_repr(dateval):
    return dateval.strftime('%Y-%m-%d %H:%M:%S %Z') if dateval else 'None'


def astimezone(obj):
    """
    Interprets an object as a timezone.

    :rtype: tzinfo

    """
    if isinstance(obj, six.string_types):
        return timezone(obj)
    if isinstance(obj, tzinfo):
        if not hasattr(obj, 'localize') or not hasattr(obj, 'normalize'):
            raise TypeError('Only timezones from the pytz library are supported')
        if obj.zone == 'local':
            raise ValueError(
                'Unable to determine the name of the local timezone -- you must explicitly '
                'specify the name of the local timezone. Please refrain from using timezones like '
                'EST to prevent problems with daylight saving time. Instead, use a locale based '
                'timezone name (such as Europe/Helsinki).')
        return obj
    if obj is not None:
        raise TypeError('Expected tzinfo, got %s instead' % obj.__class__.__name__)


def get_callable_name(func):
    """
    Returns the best available display name for the given function/callable.

    :rtype: str

    """
    # the easy case (on Python 3.3+)
    if hasattr(func, '__qualname__'):
        return func.__qualname__

    # class methods, bound and unbound methods
    f_self = getattr(func, '__self__', None) or getattr(func, 'im_self', None)
    if f_self and hasattr(func, '__name__'):
        f_class = f_self if isclass(f_self) else f_self.__class__
    else:
        f_class = getattr(func, 'im_class', None)

    if f_class and hasattr(func, '__name__'):
        return '%s.%s' % (f_class.__name__, func.__name__)

    # class or class instance
    if hasattr(func, '__call__'):
        # class
        if hasattr(func, '__name__'):
            return func.__name__

        # instance of a class with a __call__ method
        return func.__class__.__name__

    raise TypeError('Unable to determine a name for %r -- maybe it is not a callable?' % func)


def obj_to_ref(obj):
    """
    Returns the path to the given callable.

    :rtype: str
    :raises TypeError: if the given object is not callable
    :raises ValueError: if the given object is a :class:`~functools.partial`, lambda or a nested
        function

    """
    if isinstance(obj, partial):
        raise ValueError('Cannot create a reference to a partial()')

    name = get_callable_name(obj)
    if '<lambda>' in name:
        raise ValueError('Cannot create a reference to a lambda')
    if '<locals>' in name:
        raise ValueError('Cannot create a reference to a nested function')

    if ismethod(obj):
        if hasattr(obj, 'im_self') and obj.im_self:
            # bound method
            module = obj.im_self.__module__
        elif hasattr(obj, 'im_class') and obj.im_class:
            # unbound method
            module = obj.im_class.__module__
        else:
            module = obj.__module__
    else:
        module = obj.__module__
    return '%s:%s' % (module, name)


def ref_to_obj(ref):
    """
    Returns the object pointed to by ``ref``.

    :type ref: str

    """
    if not isinstance(ref, six.string_types):
        raise TypeError('References must be strings')
    if ':' not in ref:
        raise ValueError('Invalid reference')

    modulename, rest = ref.split(':', 1)
    try:
        obj = __import__(modulename, fromlist=[rest])
    except ImportError:
        raise LookupError('Error resolving reference %s: could not import module' % ref)

    try:
        for name in rest.split('.'):
            obj = getattr(obj, name)
        return obj
    except Exception:
        raise LookupError('Error resolving reference %s: error looking up object' % ref)

def safe_int(num):
    try:
        return int(num)
    except ValueError:
        result = "0"
        for c in num:
            if c not in string.digits:
                break
            result += c
        return int(result)

def bytes_to_str(b):
    if type(b) == str:
        return b
    return bytes.decode(b, "utf-8")

def bytes_to_int(b):
    if type(b) == int:
        return b
    s = bytes_to_str(b)
    return safe_int(s)

def str_to_bytes(s):
    if type(s) == bytes:
        return s
    return s.encode(encoding="utf-8")

def asint(text):
    if text is not None:
        return int(text)


# import signal
# from contextlib import contextmanager

# class TimeoutException(Exception): pass

# @contextmanager
# def time_limit(seconds):
#     def signal_handler(signum, frame):
#         raise TimeoutException("Timed out!")
#     signal.signal(signal.SIGALRM, signal_handler)
#     signal.alarm(seconds)
#     try:
#         yield
#     finally:
#         signal.alarm(0)


# import threading

# class TimeoutError(Exception):
#     pass

# class InterruptableThread(threading.Thread):
#     def __init__(self, func, *args, **kwargs):
#         threading.Thread.__init__(self)
#         self._func = func
#         self._args = args
#         self._kwargs = kwargs
#         self._result = None

#     def run(self):
#         self._result = self._func(*self._args, **self._kwargs)

#     @property
#     def result(self):
#         return self._result

# def timeout_limit(sec, f, *args, **kwargs):
#     it = InterruptableThread(f, *args, **kwargs)
#     it.start()
#     it.join(sec)
#     if not it.is_alive():
#         return it.result
#     raise TimeoutError('execution expired')




from abc import ABC,abstractmethod

DEFAULT_MAX_JOB = 10

class Tigger(ABC):
    ''''''

    def __init__(self) -> None:
        self._events = []

    @abstractmethod
    def get_next_time(self, pre):
        pass

    def is_only_once(self):
        return False


from abc import ABC,abstractmethod,abstractproperty
from contextlib import contextmanager 

class LockError(Exception):
    pass

class BaseLock(ABC):
    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def lock(self, timeout: float = None, blocking=False):
        pass

    @abstractmethod
    def unlock(self):
        pass

    @contextmanager
    def create_lock(self, timeout: float = None, blocking=False):
        succ = self.lock(timeout)
        yield succ
        if succ:
            self.unlock()


from dataclasses import dataclass
from pickle import dumps, loads

from distask.serializers.base import Serializer


@dataclass(frozen=True)
class PickleSerializer(Serializer):
    protocol: int = 4

    def serialize(self, obj) -> bytes:
        return dumps(obj, self.protocol)

    def deserialize(self, serialized: bytes):
        return loads(serialized)



import logging
import re
from typing import Optional

from distask import util
from distask.datastores.base import DataStore
from distask.serializers.base import Serializer
from distask.task import Job
from distask.util import bytes_to_int, bytes_to_str

try:
    from redis import Redis
except ImportError:
    raise ImportError('RedisDataStore requires redis installed')

class RedisDataStore(DataStore):
    '''job_id will be add group and subgroup'''
    def __init__(self, client: Redis, *, serializer: Optional[Serializer] = None,
                run_times_key: str = 'distask_times',
                jobs_key: str = 'distask_jobs', **kwargs):
            
        self._client = client
        self._serializer = serializer
        self._run_times_key = run_times_key
        self._jobs_key = jobs_key
        self._key_pattern = re.compile(r'^(\w*)::(\w*):(\w*):(\w.+)$')

    def _job_id_to_key(self, job_id):
        return self._jobs_key + "::" + job_id

    def _build_job_id(self, group, subgroup, job):
        return "{}:{}:{}".format(group, subgroup, job)

    def _build_id_key_by_job(self, job):
        job_id = self._job_id_to_key(self._build_job_id(job.group, job.subgroup, job.job_id))
        return job_id
    
    def get_jobs(self, scheduler, now, limit=None):
        job_ids = self._client.zrangebyscore(self._run_times_key, 0, now)
        if not job_ids:
            return [], None
        filter_jobs = []
        need_del_jobs = []
        has_execption = False
        for job_id in job_ids:
            job_id = bytes_to_str(job_id)
            match = self._key_pattern.match(job_id)
            if not match:
                need_del_jobs.append(job_id)
                continue
            _, group, subgroup, job = match.groups()
            if scheduler._groups and not group in scheduler._groups:
                continue

            if scheduler._subgroups and not subgroup in scheduler._subgroups:
                continue

            filter_jobs.append(job_id)

            if limit and limit > 0 and len(filter_jobs) >= limit:
                break
        jobs = []
        for job_id in filter_jobs:
            row = self._client.hgetall(job_id)
            try:
                jobs.append(self._reconstitute_job(row))
            except BaseException:
                logging.exception('Unable to restore job "%s" -- removing it', job_id)
                need_del_jobs.append(job_id)
                has_execption = True

        # Remove all the jobs we failed to restore
        if need_del_jobs:
            with self._client.pipeline() as pipe:
                pipe.delete(*need_del_jobs)
                pipe.zrem(self._run_times_key, *need_del_jobs)
                pipe.execute()
        return jobs, has_execption

    def _reconstitute_job(self, row):
        job = Job.__new__(Job)
        func_str, args = self._serializer.deserialize(row[b'call'])
        states = (
            bytes_to_str(row[b'job_id']), 
            bytes_to_int(row[b'next_time']), 
            bytes_to_str(row[b'group']), 
            bytes_to_str(row[b'subgroup']), 
            self._serializer.deserialize(row[b'tigger']),
            util.ref_to_obj(func_str), args, 
            bytes_to_int(row[b'status_last_time']),
        )
        job.__setstate__(states)
        return job

    def add_job(self, job):
        job_id = self._build_id_key_by_job(job)
        old = self._client.hgetall(job_id)
        if not old:
            document = {
                'job_id': job.job_id,
                'group': job.group,
                'subgroup': job.subgroup,
                'status_last_time': job.status_last_time,
                'next_time': job.next_time,
                'tigger': self._serializer.serialize(job.tigger),
                'call': self._serializer.serialize((job.get_func_str(), job.args)),
            }
            with self._client.pipeline() as pipe:
                pipe.multi()
                pipe.hmset(job_id, document)
                pipe.zadd(self._run_times_key, {job_id: job.next_time}) 
                pipe.execute()
        else:
            update = {
                'tigger': self._serializer.serialize(job.tigger),
                'call': self._serializer.serialize((job.get_func_str(), job.args)),
            }
            if job.next_time < bytes_to_int(old[b'next_time']):
                update['next_time'] = job.next_time
            
            with self._client.pipeline() as pipe:
                pipe.multi()
                pipe.hmset(job_id, update)
                if 'next_time' in update:
                    pipe.zadd(self._run_times_key, {job_id: update['next_time']}) 
                pipe.execute()

        return None

    def get_all_jobs(self, scheduler):
        jobs, has_execption = self.get_jobs(scheduler, util.micro_max())
        return jobs

    def modify_job(self, job, update):
        job_id = self._build_id_key_by_job(job)
        with self._client.pipeline() as pipe:
            pipe.multi()
            pipe.hmset(job_id, update)
            if 'next_time' in update:
                pipe.zadd(self._run_times_key, {job_id: update['next_time']}) 
            pipe.execute()

    def remove_job(self, job):
        job_id = self._build_id_key_by_job(job)
        
        with self._client.pipeline() as pipe:
            pipe.multi()
            pipe.delete(job_id)
            pipe.zrem(self._run_times_key, job_id) 
            pipe.execute()
        return None

    def record_job_exec(self, status, job, duration=0, runtimes=1, excetion=None):
        return None

    def clear_all_jobs_info(self):
        all_keys = self._client.keys(self._jobs_key + "*")
        with self._client.pipeline() as pipe:
            pipe.multi()
            if len(all_keys) > 0:
                pipe.delete(*all_keys)
            pipe.delete(self._run_times_key) 
            pipe.execute()


from abc import ABCMeta, abstractmethod
from base64 import b64decode, b64encode
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from re import sub
from typing import (Any, Callable, Dict, FrozenSet, Iterable, Iterator, List,
                    Optional, Set, Type)
from uuid import UUID, uuid4

from distask import util
from distask.tiggers.base import Tigger
from distask.tiggers.cron import CronTigger
from distask.tiggers.delay import DelayTigger
from distask.tiggers.interval import IntervalTigger


class DeserializationError(AttributeError):
    """Raised when a serializer fails to deserialize the given object."""

@dataclass
class Job:
    job_id: str
    next_time: int
    group: str
    subgroup: str
    tigger: Tigger = field(compare=False)
    func: Callable = field(compare=False)
    args: tuple = field(compare=False)
    status_last_time: int
    deal_max_time: int = field(init=False, default=60_000)
    close_now: bool = field(init=False, default=False)

    def __init__(self, func, tigger, args=None, job_id=None, next_time=None, group='', subgroup='', status_last_time=0, **kwargs):
        
        if tigger == "interval":
            self.tigger = IntervalTigger(**kwargs)
        elif tigger == "delay":
            self.tigger = DelayTigger(**kwargs)
        elif tigger == "cron":
            self.tigger = CronTigger(**kwargs)
        elif issubclass(tigger, Tigger):
            self.tigger = tigger
        else:
            raise AttributeError("unknow target({})".format(tigger))

        only_once = self.tigger.is_only_once()
        self.func = func
        self.args = args
        if job_id == None:
            if only_once:
                self.job_id = f"{func.__module__}.{func.__name__}__" + "{}".format(uuid4())
            else:
                self.job_id = f"{func.__module__}.{func.__name__}"
        else:
            self.job_id = job_id

        self.next_time = self.tigger.get_next_time()[0]
        self.group = group
        self.subgroup = subgroup
        self.status_last_time = status_last_time

    def get_next_time(self):
        return self.tigger.get_next_time(self.next_time)

    def call_func(self, tiggers=None, scheduler=None):
        if not tiggers: tiggers = self.get_next_time()
        is_close = self.func(tiggers, *self.args, self.group, self.subgroup, scheduler=scheduler)
        return tiggers[-1], is_close

    def is_only_once(self):
        return self.tigger.is_only_once()

    def get_func_str(self):
        return util.obj_to_ref(self.func)

    def set_func_str(self, func_str):
        self.func = util.ref_to_obj(func_str)

    def __getstate__(self):  
        """Return state values to be pickled."""
        return (self.job_id, self.next_time, self.group, self.subgroup, self.tigger, self.func, self.args, self.status_last_time)  

    def __setstate__(self, state):
        """Restore state from the unpickled state values."""  
        self.job_id, self.next_time, self.group, self.subgroup, self.tigger, self.func, self.args, self.status_last_time = state



from distask.serializers.json import JSONSerializer
from distask.serializers.pickle import PickleSerializer
from distask.locks.base import BaseLock
from distask.datastores.base import DataStore
from distask.schedulers.base import Scheduler
from distask.schedulers.background import BackgroundScheduler
from distask.tiggers.cron import CronTigger
from distask.tiggers.delay import DelayTigger
from distask.tiggers.interval import IntervalTigger
from distask.task import Job

def register_job(scheduler: Scheduler, *args, **kwargs) -> callable:
    def wrapper_register_job(func):
        job = Job(func, *args, **kwargs)
        scheduler.add_job(job)
        return func
    return wrapper_register_job

def create_scheduler(client_data, lock_data, serialize="pickle", backgroud=False, **kwargs):
    serialize =  JSONSerializer() if serialize == "json" else PickleSerializer()
    client = None
    store = None
    if isinstance(client_data, DataStore):
        store = client_data
    else:
        cliet_type = client_data.pop("t")
        if cliet_type == "mongo":
            try:
                from pymongo import MongoClient
                from distask.datastores.mongodb import MongoDataStore
                client_args = client_data.pop("client_args", {})
                store_args = client_data.pop("store_args", {})
                client = MongoClient(**client_args)
                store = MongoDataStore(client, serializer=serialize, **store_args)
            except:
                raise
        elif cliet_type == "redis":
            try:
                from redis import Redis
                from distask.datastores.redis import RedisDataStore
                client_args = client_data.pop("client_args", {})
                store_args = client_data.pop("store_args", {})
                client = Redis(**client_args)
                store = RedisDataStore(client, serializer=serialize, **store_args)
            except:
                raise
            
    assert store, "must store client exist"
    lock = None
    if isinstance(lock_data, BaseLock):
        lock = lock_data
    else:
        lock_type = lock_data.pop("t")
        if lock_type == "rllock":
            try:
                from distask.locks.rl_lock import RLLock
                args = lock_data.pop("args", [])
                lock = RLLock(*args, **lock_data)
            except:
                raise
        if lock_type == "zklock":
            try:
                from distask.locks.zk_lock import ZKLock
                args = lock_data.pop("args", [])
                lock = ZKLock(*args, **lock_data)
            except:
                raise

    assert lock, "must lock exist"

    if backgroud:
        scheduler = BackgroundScheduler(store=store, lock=lock, **kwargs)
    else:
        scheduler = Scheduler(store=store, lock=lock, **kwargs)
    return scheduler